@@ -3,7 +3,7 @@
   | Suhosin Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2006-2007 The Hardened-PHP Project                     |
-  | Copyright (c) 2007 SektionEins GmbH                                  |
+  | Copyright (c) 2007-2010 SektionEins GmbH                             |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
@@ -32,7 +32,11 @@
 #include "SAPI.h"
 #include "php_variables.h"
 
+#if PHP_VERSION_ID >= 50300
+static int (*orig_header_handler)(sapi_header_struct *sapi_header, sapi_header_op_enum op, sapi_headers_struct *sapi_headers TSRMLS_DC) = NULL;
+#else
 static int (*orig_header_handler)(sapi_header_struct *sapi_header, sapi_headers_struct *sapi_headers TSRMLS_DC) = NULL;
+#endif
 
 char *suhosin_encrypt_single_cookie(char *name, int name_len, char *value, int value_len, char *key TSRMLS_DC)
 {
@@ -221,10 +225,20 @@ char *suhosin_cookie_decryptor(TSRMLS_D)
 
 /* {{{ suhosin_header_handler
  */
+#if PHP_VERSION_ID >= 50300
+int suhosin_header_handler(sapi_header_struct *sapi_header, sapi_header_op_enum op, sapi_headers_struct *sapi_headers TSRMLS_DC)
+#else
 int suhosin_header_handler(sapi_header_struct *sapi_header, sapi_headers_struct *sapi_headers TSRMLS_DC)
+#endif
 {
 	int retval = SAPI_HEADER_ADD, i;
 	char *tmp;
+
+#if PHP_VERSION_ID >= 50300
+	if (op != SAPI_HEADER_ADD && op != SAPI_HEADER_REPLACE) {
+                goto suhosin_skip_header_handling;
+	}
+#endif
 	
 	if (!SUHOSIN_G(allow_multiheader) && sapi_header && sapi_header->header) {
 	
@@ -309,10 +323,14 @@ int suhosin_header_handler(sapi_header_struct *sapi_header, sapi_headers_struct
 		sapi_header->header_len = len;
 	}
 
-
+suhosin_skip_header_handling:
 	/* If existing call the sapi header handler */
 	if (orig_header_handler) {
+#if PHP_VERSION_ID >= 50300
+		retval = orig_header_handler(sapi_header, op, sapi_headers TSRMLS_CC);
+#else
 		retval = orig_header_handler(sapi_header, sapi_headers TSRMLS_CC);
+#endif
 	}
 
 	return retval;
